use vulkano::buffer::BufferContents;
use vulkano::buffer::{BufferCreateInfo, BufferUsage};
use vulkano::memory::allocator::{AllocationCreateInfo, MemoryTypeFilter};
use vulkano::pipeline::graphics::vertex_input::Vertex;

use crate::core::render::primitives::{AsBindableBuffer, AsIndexBuffer, AsVertexBuffer};

#[derive(BufferContents, Vertex, Clone, Copy)]
#[repr(C)]
pub struct Vertex2D {
    #[format(R32G32_SFLOAT)]
    pub position: [f32; 2],

    #[format(R32G32_SFLOAT)]
    pub uv: [f32; 2],
}

#[derive(BufferContents, Vertex, Clone, Copy)]
#[repr(C)]
pub struct Vertex3D {
    #[format(R32G32B32_SFLOAT)]
    pub position: [f32; 3],

    #[format(R32G32_SFLOAT)]
    pub uv: [f32; 2],
}

impl AsBindableBuffer for Vertex2D {
    type BufferData = Vertex2D;

    fn buffer_create_info() -> BufferCreateInfo {
        BufferCreateInfo {
            usage: BufferUsage::VERTEX_BUFFER,
            ..Default::default()
        }
    }

    fn allocation_create_info() -> AllocationCreateInfo {
        AllocationCreateInfo {
            memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
            ..Default::default()
        }
    }

    fn to_buffer_data(&self) -> Self::BufferData {
        *self
    }
}

impl AsVertexBuffer for Vertex2D {}

impl AsBindableBuffer for Vertex3D {
    type BufferData = Vertex3D;

    fn buffer_create_info() -> BufferCreateInfo {
        BufferCreateInfo {
            usage: BufferUsage::VERTEX_BUFFER,
            ..Default::default()
        }
    }

    fn allocation_create_info() -> AllocationCreateInfo {
        AllocationCreateInfo {
            memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
            ..Default::default()
        }
    }

    fn to_buffer_data(&self) -> Self::BufferData {
        *self
    }
}

impl AsVertexBuffer for Vertex3D {}

impl AsBindableBuffer for u32 {
    type BufferData = u32;

    fn buffer_create_info() -> BufferCreateInfo {
        BufferCreateInfo {
            usage: BufferUsage::INDEX_BUFFER,
            ..Default::default()
        }
    }

    fn allocation_create_info() -> AllocationCreateInfo {
        AllocationCreateInfo {
            memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
            ..Default::default()
        }
    }

    fn to_buffer_data(&self) -> Self::BufferData {
        *self
    }
}

impl AsIndexBuffer for u32 {}
